using namespace std;

#include <iostream>
#include <math.h>
#include <stdlib.h>
#include <time.h>

#include "bitset.h"

/* number of input world-set descriptors */
int NUM_WSDS;

/* size of input world-set descriptors */
int WSD_LEN;


/* number of variables */
int NUM_VARS; 

/*
 * max domain size of variables (the timing reported in
 * the paper uses the same domain size for all variables)
 */
int RNG_MAX; 

/* the mutex computation uses one world-set descriptor, whose size can
   increase during computation. MAX_LEN is the maximum number of
   assignment entries allocated for this world-set descriptor.
 */
int MAX_LEN;

/* if in test mode, omit printing sth so that timing (done by an
 * external tool like /usr/bin/time) can be directly collected into
 * gnuplot files 
 */
int TEST_MODE;

/* if in print mode, just print the generated data (variables, dom
 * size, world-set descriptors) and exit.
 */
int PRINT_MODE;


/* which algorithm should be used: ve (variable elimination), indve
   (ve+independent partitioning), we (world-set descriptor
   elimination), wve (wsd+var elimination).
 */
int WHICH_ALGO;

/* in case of variable elimination we can now use one of two
 * heuristics: minlog and minmax, as detailed in the paper.
 */
int WHICH_HEURISTIC;

/* which data generation should be used: 0 (random), 1 (random from
 * joined BIDs), 2 (random from self-joined BIDs)
 */
int DATA_GEN;


/*
 * the variable var occurs (in a given set of world-set descriptors)
 * no_occurrences many times (possible with different assignments).
 */
typedef struct {
	int var;
	int no_occurrences;
} occurrence;


/*
 * Map = var assignment var->rng with probability prob. we also keep
 * reference to the world-set descriptor (wsd_host) using this
 * mapping. neg is only used in case of mutex computation to
 * succinctly represent as not(var->rng) all other assignments of
 * variable var.
 * LIMITATION: we assume uniform prob distribution of values for each var.
 */
typedef struct 
{
  int var;
  int rng;
  int neg;
  double prob; // of var->rng; if neg=1, then the real prob is 1-prob.
  int wsd_host;
} Map;



/* 
 * A world-set decriptor is an array of assignments of different vars.
 * its probability is the product of prob of its var assignments.
 */
typedef struct
{
  Map** data;
  double prob;
} WSD;


int** var_rng;
WSD** S;
int* Rng;



/***************** HEURISTICS *************************/

/* find (v,n) = the variable with max occurrences (and occurrence number) in the given wsd set */
occurrence*
find_var_maxoccur(bitset* set)
{
  int occ[NUM_VARS];
  memset(occ, 0, NUM_VARS * sizeof(int));

  for(int i = 0; i < NUM_WSDS; i++)
    if (bitset_test_bit(set, i))
      for (int j = 0; j < WSD_LEN; j++)
	occ[S[i]->data[j]->var]++;

  occurrence* o = (occurrence*) malloc(sizeof(occurrence));
  o->var = 0;
  o->no_occurrences = occ[0];

  for (int i = 1; i < NUM_VARS; i++)
    if (occ[i] > o->no_occurrences) {
      o->no_occurrences = occ[i];
      o->var = i;
    }

  return o;
}

/* if a variable occurs sufficiently many times, then use DP (return
 * var), otherwise non-overlapping computation (return -1).
 */
int
HEURISTIC1(bitset* subset, int subset_size) 
{
  double threshold = 0;

  occurrence* o  = find_var_maxoccur(subset);
  
  if (o->no_occurrences > threshold * subset_size)
    return o->var;

  return -1;
}


/* find var with best partitioning of the given wsd set */
int
choose_var_minlog(bitset* set)
{
  for (int i = 0; i < NUM_VARS; i++)
    for (int j = 0; j < Rng[i]; j++)
      var_rng[i][j] = 0;

  int set_size = 0;

  for(int i = 0; i < NUM_WSDS; i++)
    if (bitset_test_bit(set, i)) {
      for (int j = 0; j < WSD_LEN; j++) {
	int var = S[i]->data[j]->var;
	var_rng[var][S[i]->data[j]->rng]++;
      }
      set_size++;
    }

  int var = -1;
  double val = exp(1)*set_size;

  for (int i = 0; i < NUM_VARS; i++) {

    int how_many_diff_assignments = 0;
    int how_many_assignments = 0;

    for (int j = 0; j < Rng[i]; j++)
      if(var_rng[i][j] > 0) {
	how_many_diff_assignments++;
	how_many_assignments += var_rng[i][j];
      }

    if (how_many_diff_assignments != 0) {
      int t = set_size - how_many_assignments;
      double new_val = 0.0;
    
      if (how_many_diff_assignments != Rng[i]) 
	new_val = t;

      for (int j = 0; j < Rng[i]; j++)
	if(var_rng[i][j] > 0)
	  new_val += log(1 + exp(var_rng[i][j]+t - new_val));
      
      if (val > new_val) {
	val = new_val;
	var = i;
      }
    }
  }

  return var;
}

/*
 * estimate = ln(e^s1+..+e^s_n), where si>0. si=size of partition for
 * an assignment of var that occurs in wsd set plus size of wsd subset
 * without var; there is an s_i for size of wsd subset without var.
 *
 * rec. formula for computing the estimate: 
 * f_k = f_{k-1}+ln(1+e^{sk-f_{k-1}}
 */
int
minlog_estimate(bitset* set) 
{
  return choose_var_minlog(set);
}

/* find var with min max occurrences of an assignment*/
int
choose_var_minmax(bitset* set)
{
  for (int i = 0; i < NUM_VARS; i++)
    for (int j = 0; j < Rng[i]; j++)
      var_rng[i][j] = 0;

  for(int i = 0; i < NUM_WSDS; i++)
    if (bitset_test_bit(set, i))
      for (int j = 0; j < WSD_LEN; j++)
	var_rng[S[i]->data[j]->var][S[i]->data[j]->rng]++;

  int var = -1;
  int max_assignments_all = NUM_WSDS;

  for (int i = 0; i < NUM_VARS; i++) {
      int max_assignments_var = 0;
      for (int j = 0; j < Rng[i]; j++)
	if(var_rng[i][j] > max_assignments_var)
	  max_assignments_var = var_rng[i][j];

      if (max_assignments_var != 0 && max_assignments_var < max_assignments_all) {
	max_assignments_all = max_assignments_var;
	var = i;
      }
  }

  return var;
}

/*
 * estimate = max(s1,..,s_n), where si>0. si=size of partition for
 * an assignment of var that occurs in wsd set plus size of wsd subset
 * without var; there is an s_i for size of wsd subset without var.
 *
 * choose a var with a minimal estimate
 */
int
minmax_estimate(bitset* set) 
{
  return choose_var_minmax(set);
}



/**** CREATE/MANIPULATE WORLD-TABLE, MAPS, & WSDs ****************************/


//define world-table
void
create_world_table ()
{
  if (PRINT_MODE)
    cout << " Vars 0 through " << NUM_VARS << " have domain size " << RNG_MAX << " and prob " << 1.0/RNG_MAX << endl;
  for (int i=0; i<NUM_VARS; i++) 
    Rng[i] = RNG_MAX;
}

void
create_var_rng_table ()
{
  var_rng = (int**) malloc (NUM_VARS * sizeof(int*));
  for (int i=0; i<NUM_VARS; i++) 
    var_rng[i] = (int*) malloc (Rng[i] * sizeof(int));
}


int
var_already_used(int var, Map** maps, int j)
{
  if (maps)
    for (int i = 0; i < j; i++)
      if (maps[i] && maps[i]->var == var)
	return 1;
  return 0;
}

void
print_map (Map* m, int with_prob)
{
  cout << "(";
  if (m->neg)
    cout << "not ";
  cout << m->var << "->" << m->rng;
  if (with_prob)
    cout  << "[" << m->prob << "]";
  cout << ")";
}

void
print_wsd (WSD* wsd, int with_prob)
{
  cout << "[";
  for (int i=0; i<WSD_LEN; i++)
    {
      print_map(wsd->data[i], 0);
      if (i<WSD_LEN-1) cout << ",";
    }
  cout << "]";
  if (with_prob)
    cout << "(" << wsd->prob << ")";
  cout << endl;
}


void
print_maps (Map** wsd, int length)
{
  cout << "[";
  for (int i=0; i<length; i++)
    {
      if (wsd[i] != NULL)
	print_map(wsd[i], 0); cout << " {" << wsd[i]->wsd_host << "}";
      if (i<length-1) cout << ",";
    }
  cout << "]" << endl;
}


Map* 
create_map (int var, int rng, int neg, double prob, int wsd_host)
{
  Map* m      = (Map*) malloc (sizeof(Map));
  m->var      = var;
  m->rng      = rng;
  m->neg      = neg;
  m->prob     = prob;
  m->wsd_host = wsd_host;
  return m;
}

Map* 
copy_neg_map (Map* map)
{
  Map* m_copy = (Map*)malloc(sizeof(Map));
  m_copy->var = map->var;
  m_copy->rng = map->rng;
  m_copy->neg = 1 - map->neg;
  m_copy->prob = map->prob;
  m_copy->wsd_host = map->wsd_host;
  
  return m_copy;
}


/* returns
 * -2: equal 
 * -1: mutex 
 * 0: different vars
 * 2: same vars, diff rng, both neg
 * 3: same vars, diff rng, diff neg (the positive one constrains the negative one)
 */
int
compare_map (Map* m1, Map* m2)
{
  if (!m1 || !m2 || m1->rng == -1 || m2->rng == -1)
    return 0;

  if (m1->var == m2->var)
    if (m1->rng == m2->rng)
      if (m1->neg == m2->neg)
	return -2;
      else
	return -1;
    else
      if (m1->neg != m2->neg)
	return 3;
      else
	if (m1->neg && m2->neg)
	  return 2;
	else 
	  return -1;
  return 0;
}


/*
 * 0 if not mutex and 1 if mutex.
 */
int
wsd_mutex (Map** d1, int len1, Map** d2, int len2)
{
  int is_mutex = 0;

  int i=0;
  while (i<len1 && !is_mutex)
    {
      int j=0;
      while (j<len2 && compare_map(d1[i],d2[j])!=-1)
	j++;
      if (j<len2)
	is_mutex = 1;
      i++;
    }
  return is_mutex;
}

/*
 * 0 if not independent and 1 if independent.
 */
int 
wsd_ind (WSD* d1, WSD* d2)
{
  int is_independent = 1;

  int i=0;
  while (i<WSD_LEN && is_independent)
    {
      int j=0;
      while (j<WSD_LEN && compare_map(d1->data[i],d2->data[j]) == 0)
        j++;
      if (j<WSD_LEN)
        is_independent = 0;
      i++;
    }
  return is_independent;
  
}





int
remove_maps_host (Map** wsd, int length, int wsd_host)
{
  int j = length-1;
  while (j > -1 && wsd[j]->wsd_host == wsd_host) {
    if (j >= WSD_LEN) {// remove the newly created objects only
      free (wsd[j]);
      wsd[j] = NULL;
    } else break;
    j--;
  }
  return j+1;
}

void
reset_wsd (Map** wsd, int length)
{
  for (int i = 0; i < length; i++)
    {
      if (i > WSD_LEN) // remove the created objects
	free (wsd[i]);
      wsd[i] = NULL;
    }
}

int
count_negvar (Map** wsd, int length, int var)
{
  int count = 0;

  for (int i = 0; i < length; i++) 
    if (wsd[i]->neg && wsd[i]->var == var)
      count++;

  return count;
}

Map*
find_equal_map (Map** wsd, int length, Map* map)
{
  for (int i = 0; i < length; i++)
    if (compare_map(wsd[i], map) == -2)
      return (wsd[i]);
  return (NULL);
}

void
create_random_WSDs ()
{
  if (PRINT_MODE)
    cout << " ********\t WSDs\t ********* " << endl;

  for (int i=0; i<NUM_WSDS; i++)
    {
      S[i] = (WSD*) malloc(sizeof(WSD));
      S[i]->data = (Map**)malloc(WSD_LEN*sizeof(Map*));
      S[i]->prob = 1.0;
      for (int j=0; j<WSD_LEN; j++)
	{
	  int var;
	  int rng;
	  do {
	    var = (int) (NUM_VARS * (rand() / (RAND_MAX + 1.0)));
	  } while (var_already_used(var, S[i]->data, j));
	  rng = (int) (Rng[var] * (rand() / (RAND_MAX + 1.0)));
	  S[i]->data[j] = create_map (var, rng, 0, 1.0/Rng[var], i);
	  S[i]->prob = S[i]->prob * (1.0/Rng[var]);
	}

      if (PRINT_MODE) {
	cout << i << "\t"; 
	print_wsd(S[i],0);
      }
    }
}

/* wsds are generated via queries (without self-joins) on BIDs.
 * simulation of query eval on BIDs: 
 * partition the var set into WSD_LEN partitions (each partition=BID, WSD_LEN-1 joins)
 * valid wsds do not only mappings from the same partition
 * WSD_LEN is minimum 2
 */

void
create_random_WSDs_from_BIDs ()
{
  if (PRINT_MODE)
    cout << " ********\t WSDs\t ********* " << endl;

  S = (WSD**)malloc(NUM_WSDS*sizeof(WSD*));

  int var_partitions[WSD_LEN];
  int count = 0;

  for (int i = 0; i < WSD_LEN-1; i++) {
    var_partitions[i] = 1+(int) (NUM_VARS * (rand() / (WSD_LEN*(RAND_MAX + 1.0))));
    count += var_partitions[i];
  }

  var_partitions[WSD_LEN-1] = NUM_VARS - count;

  for (int i=0; i<NUM_WSDS; i++)
    {
      S[i] = (WSD*) malloc(sizeof(WSD));
      S[i]->data = (Map**)malloc(WSD_LEN*sizeof(Map*));
      S[i]->prob = 1.0;
      count = 0;
      for (int j=0; j<WSD_LEN; j++)
	{
	  int var = count + (int) (var_partitions[j] * (rand() / (RAND_MAX + 1.0)));
	  count += var_partitions[j];

	  int rng = (int) (Rng[var] * (rand() / (RAND_MAX + 1.0)));
	  S[i]->data[j] = create_map (var, rng, 0, 1.0/Rng[var], i);
	  S[i]->prob = S[i]->prob * (1.0/Rng[var]);
	}
      
      if (PRINT_MODE != 0) {
	cout << i << "\t"; 
	print_wsd(S[i],0);
      }
    }
}

/* wsds are generated via queries (with self-joins) on BIDs.
 * simulation of query eval on BIDs: 
 * partition the var set into WSD_LEN partitions (each partition=BID, WSD_LEN-1 joins)
 * valid wsds do not only mappings from the same partition
 * WSD_LEN is minimum 2
 */

void
create_WSDs_from_selfjoined_BIDs (double join_selectivity)
{
  if (PRINT_MODE)
    cout << " ********\t WSDs\t ********* " << endl;

  if (WSD_LEN < 2) {// we must do the self-join
    cout << "wsd length is too small:" << 2 << endl;
    exit(-1);
  }

  int var_partitions[WSD_LEN];
  int count = 0;

  NUM_WSDS = 1;

  for (int i = 0; i < WSD_LEN-1; i++) {
    var_partitions[i] = 2+(int) (NUM_VARS * (rand() / (WSD_LEN*(RAND_MAX + 1.0))));
    int BID_size = 0;
    for (int j = count; j < count+var_partitions[i]; j++)
      BID_size += Rng[j];
    if (i == 0)
      NUM_WSDS = (int)(BID_size * join_selectivity * BID_size);
    else
      NUM_WSDS = (int)(NUM_WSDS * join_selectivity * BID_size);
      
    count += var_partitions[i];
  }

  if (PRINT_MODE)
    cout << "num wsds: " << NUM_WSDS << endl;

  var_partitions[WSD_LEN-1] = NUM_VARS - count;

  //self join (partition[0]), then WSD_LEN-2 joins on diff tables (partition[1]..partition[WSd_LEN-2])

  S = (WSD**)malloc(NUM_WSDS*sizeof(WSD*));
  for (int i=0; i<NUM_WSDS; i++)
    {
      S[i] = (WSD*) malloc(sizeof(WSD));
      S[i]->data = (Map**)malloc(WSD_LEN*sizeof(Map*));
      S[i]->prob = 1.0;

      int var = (int) (var_partitions[0] * (rand() / (RAND_MAX + 1.0)));
      int rng = (int) (Rng[var] * (rand() / (RAND_MAX + 1.0)));
      S[i]->data[0] = create_map (var, rng, 0, 1.0/Rng[var], i);
      
     int var2;
      do
	var2 = (int) (var_partitions[0] * (rand() / (RAND_MAX + 1.0)));
      while (var2 == var);

      rng = (int) (Rng[var] * (rand() / (RAND_MAX + 1.0)));
      S[i]->data[1] = create_map (var2, rng, 0, 1.0/Rng[var], i);
 
      count = var_partitions[0];
      for (int j=2; j<WSD_LEN; j++)
	{
	  int var = count + (int) (var_partitions[j] * (rand() / (RAND_MAX + 1.0)));
	  count += var_partitions[j];

	  int rng = (int) (Rng[var] * (rand() / (RAND_MAX + 1.0)));
	  S[i]->data[j] = create_map (var, rng, 0, 1.0/Rng[var], i);
	  S[i]->prob = S[i]->prob * (1.0/Rng[var]);
	}
      
      if (PRINT_MODE) {
	cout << i << "\t"; 
	print_wsd(S[i],0);
      }
    }
}



/******************************************/



// assumption: all maps of the wsd are unique
double
probability(Map ** wsd, int length)
{
  double p = 1.0;
  int i, j;
  int processed[length];

  for (i = 0; i<length; i++)
    processed[i] = 0;

  i = 0;
  while (i < length)
    {
      if (!wsd[i]->neg) {
	p = p * wsd[i]->prob;
      } else
	if (!processed[i]) {
	  double _p = 1 - wsd[i]->prob;
	  j = i+1;
	  while (j < length)
	    {
	      if (wsd[j]->var == wsd[i]->var && wsd[j]->neg) {
		_p -= wsd[j]->prob;
		processed[j] = 1;
	      }
	      j++;
	    }
	  p *= _p;
	}
    i++;
    }

  return p;
}


/* return the subset of wsds that are mutex of the rest */
bitset*
find_mutex_wsds(bitset* set)
{
  bitset* mutex_wsds = bitset_init(NUM_WSDS);
  bitset_reset(mutex_wsds);
  int no_mutex = 1;

  for (int k = 0; k < NUM_WSDS; k++)
      if (bitset_test_bit(set,k)) {
	int j = 0;
	while (j < NUM_WSDS && 
	       (j==k || bitset_test_bit(set,j) && wsd_mutex(S[k]->data, WSD_LEN, S[j]->data, WSD_LEN)))
	  j++;

	if (j == NUM_WSDS) {//S[k] is mutex with the rest
	  bitset_set_bit(mutex_wsds,k);
	  no_mutex = 0;
	}
      }
  if (no_mutex) {
    bitset_free(mutex_wsds);
    return NULL;
  }
  return mutex_wsds;
}


int
minmax_estimate_mutex(bitset* set)
{
  int n_mutex[NUM_WSDS];
  memset(n_mutex, 0, NUM_WSDS * sizeof(int));

  int set_size = 0;
  for (int k = 0; k < NUM_WSDS; k++)
    if (bitset_test_bit(set,k)) {
      for (int j=k+1; j<NUM_WSDS; j++)
	if (bitset_test_bit(set,j) && wsd_mutex(S[k]->data, WSD_LEN, S[j]->data, WSD_LEN)) {
	  n_mutex[j]++;
	  n_mutex[k]++;
	}
      set_size++;
    }

  int max_mutex = 0;
  int pos = -1;
  for (int k = 0; k < NUM_WSDS; k++)
    if (bitset_test_bit(set,k) && max_mutex < n_mutex[k]) {
      max_mutex = n_mutex[k];
      pos = k;
    }

  if (max_mutex != set_size-1)
    pos = -1;

  return pos;
}


void
dfs (bitset* set, bitset* set_to_construct, int idx)
{
  bitset_set_bit(set_to_construct,idx);
  for (int j=0; j<NUM_WSDS; j++)
    if (bitset_test_bit(set,j) && !bitset_test_bit(set_to_construct,j) &&
    	!wsd_ind(S[idx],S[j]) && idx != j)
      dfs(set,set_to_construct,j);
}

/******************************************/

/* partition the given wsd set into 2 independent sets
 * first set is returned and contains dependent wsds.
 */
bitset*
find_independent_split(bitset* set)
{
  if (bitset_test_empty(set))
    return NULL;

  bitset *subset  = bitset_init(NUM_WSDS);
  bitset_reset(subset);

  int i = 0;
  while (i < NUM_WSDS && !bitset_test_bit(set,i))
    i++;
  if (i == NUM_WSDS) //oops, set empty. WRONG.
    exit(-1);

  dfs(set,subset,i);
  
  return subset;
}


//the temporarily dropped mappings (see find_wsds_with_var_rng) are restored.
void
reset_wsds_var_rng(bitset* set, int var, int rng)
{
  for (int i = 0; i < NUM_WSDS; i++)
    if (bitset_test_bit(set,i))
      for (int j = 0; j < WSD_LEN; j++)
	if(S[i]->data[j]->var == var) {
	  S[i]->data[j]->rng = rng;
	  S[i]->prob *= S[i]->data[j]->prob;
	}
  
}


//find the wsds that contain mapping var->rng; also *temporarily* drop this mapping.
bitset*
find_wsds_with_var_rng(bitset* set, int var, int rng)
{
  bitset *subset  = NULL;
  int found_empty_wsd = 0;

  int i = 0;
  while (i < NUM_WSDS && !found_empty_wsd) {
    if (bitset_test_bit(set,i)) {
      int j = 0;
      while (j < WSD_LEN && (S[i]->data[j]->var != var || S[i]->data[j]->rng != rng))
	j++;
      if (j < WSD_LEN) {
	if (subset == NULL) {
	  subset = bitset_init(NUM_WSDS);
	  bitset_reset(subset);
	}
	bitset_set_bit(subset,i);

	//mark the used choices of rng for that var.
	S[i]->data[j]->rng = -1;
	S[i]->prob /= S[i]->data[j]->prob;

	//03.2008: check if the remaining S[i] is empty, 
	//in which case the prob of all wsds agreeing with (var,rng) is just S[i]->data[j]->prob :-)

	if (S[i]->prob == 1.0)
	  found_empty_wsd = 1;
      }
    }
    i++;
  }
  if (found_empty_wsd) {
    reset_wsds_var_rng(subset, var, rng);
    bitset_reset(subset);
  }

  return subset;
}



//find the wsds that do not contain mapping of the given var
bitset*
find_wsds_without_var(bitset* set, int var)
{
  bitset *subset  = NULL;

  for (int i = 0; i < NUM_WSDS; i++)
    if (bitset_test_bit(set,i)) {
      int j = 0;
      while (j < WSD_LEN && S[i]->data[j]->var != var)
	j++;
      if (j == WSD_LEN) {
	if (subset == NULL) {
	  subset = bitset_init(NUM_WSDS);
	  bitset_reset(subset);
	}
	bitset_set_bit(subset,i);
      }
    }
  
  return subset;
}


/*
 * compute the probability of a given set of wsds using independent
 * partitioning and variable elimination; set is the subset of wsd
 * S set for which to compute probability
 */
double
indve_compute_prob (bitset* set)
{
  double p_left = 0.0;
  double p_right = 0.0;

  bitset *subset;

  if (bitset_test_empty(set))
    return 0.0;

  //find subset of S independent of the rest (subset repr. dependent wsds)
  subset = find_independent_split(set);
 
  /********* process the left subset containing dependent wsds ***************/
  
  int pos = bitset_test_singleton(subset);
  if (pos != -1) //special case of 1 wsd
    p_left = S[pos]->prob;
  else { //subset contains more than one wsd

  /******************* PLACE TO PLUG IN HEURISTICS **********************/

    int var;
    if (!WHICH_HEURISTIC)
      var = minlog_estimate(subset);
    else
      var = minmax_estimate(subset);
    if (var == -1) // all vars are used in the wsd set
      return 1.0;
    
    bitset* subset_without_var = find_wsds_without_var(subset, var);
    
    double p_without_var = -1;
    
    for (int i = 0; i < Rng[var]; i++) {
      bitset* subset_var_rng = find_wsds_with_var_rng(subset, var, i);
  
      if (subset_var_rng == NULL) {
	if (p_without_var == -1)
	  p_without_var = indve_compute_prob (subset_without_var);
	p_left += p_without_var;
      }
      else {
	if (bitset_test_empty(subset_var_rng))
	  p_left += 1.0/Rng[var];
	else {
	  bitset_union(subset_var_rng, subset_without_var);
	  p_left += indve_compute_prob (subset_var_rng);
	  reset_wsds_var_rng(subset_var_rng, var, i);
	}
      }
      if (subset_var_rng)
	bitset_free(subset_var_rng);

#ifndef CONDITIONING
      if ((p_left == 1.0) && (Rng[var] == 1)) {//stop early
	if (subset_without_var)
	  bitset_free(subset_without_var);
	bitset_free(subset);
	return 1.0;
      }
#endif
    }
    if (subset_without_var)
      bitset_free(subset_without_var);
    p_left = p_left * (1.0/Rng[var]);
  }

  if (p_left == 1.0) {// stop early
    bitset_free(subset);
    return 1.0;
  }

  /********* process recursively the right subset ***************/


  bitset_negate(set,subset);
  p_right = indve_compute_prob(subset);


  bitset_free(subset);

  /******* combine the probabilities of left and right subsets ***************/

  return p_left + p_right - p_left*p_right;
}


/*
 * compute the probability of a given set of wsds using variable
 * elimination; set states the subset of wsd S set for which
 * to compute confidence
 */
double
ve_compute_prob (bitset* set)
{
  double prob = 0.0;

  if (bitset_test_empty(set))
    return 0.0;

  int pos = bitset_test_singleton(set);
  if (pos != -1) //special case of 1 wsd
    return S[pos]->prob;
  
  //set contains more than one wsd

  /******************* PLACE TO PLUG IN HEURISTIC **********************/
  
  int var = -1;
  if (!WHICH_HEURISTIC)
    var = minlog_estimate(set);
  else
    var = minmax_estimate(set);
  if (var == -1) // all vars are used in the wsd set
    return 1.0;
  
  bitset* subset_without_var = find_wsds_without_var(set, var);
  
  double p_without_var = -1;
  
  for (int i = 0; i < Rng[var]; i++) {
    bitset* subset_var_rng = find_wsds_with_var_rng(set, var, i);
    if (subset_var_rng == NULL) {
      if (p_without_var == -1)
	p_without_var = ve_compute_prob (subset_without_var);
      prob += p_without_var;
    }
    else {
      if (bitset_test_empty(subset_var_rng))
	prob += 1.0/Rng[var];
      else {
	bitset_union(subset_var_rng, subset_without_var);
	prob += ve_compute_prob (subset_var_rng);
	reset_wsds_var_rng(subset_var_rng, var, i);
      }
    }
    if (subset_var_rng)
      bitset_free(subset_var_rng);
    if (prob/Rng[var] == 1.0) {//stop early
      if (subset_without_var)
	bitset_free(subset_without_var);
      return 1.0;
    }
  }
  if (subset_without_var)
    bitset_free(subset_without_var);

  return prob * (1.0/Rng[var]);
}

/* computes prob of wsd set by iterating through all mutex wsds, 
 * whose set is equivalent to the given set.
 * assumption: set is not empty
 */
double
we_compute_prob(bitset* set)
{
  Map** crtWSD = (Map**)malloc(WSD_LEN * sizeof(Map*)); //possibly extended later
  MAX_LEN = WSD_LEN;
  int crt_len;
  double prob = 0.0;

  //for statistics only
  //  int countWSDs = 0;
  //  int sizeWSDs  = 0;

  for (int k = 0; k < NUM_WSDS; k++)
    {
      if (bitset_test_bit(set,k)) {
	//initialize crtWSD with S[k]
	for (int i=0; i<WSD_LEN; i++)
	  crtWSD[i] = S[k]->data[i];
	crt_len = WSD_LEN;
	//print_maps(crtWSD,crt_len); cin.get();

	int diff[NUM_WSDS];
	int dlen = 0;
	for (int j=k+1; j<NUM_WSDS; j++)
	  {
	    if (bitset_test_bit(set,j) && !wsd_mutex(crtWSD, crt_len, S[j]->data, WSD_LEN)) {
	      dlen++;
	      diff[dlen-1] = j;
	    }
	  }
	
	int pos[dlen];
	for (int i=0; i<dlen; i++)
	  pos[i] = 0;
	int crt_pos = 0;
	
	
	if (dlen == 0) {
	    //crtWSD[k] is mutex with the rest.
	    double p = probability(crtWSD, crt_len);
	    prob += p;
	    //countWSDs++; sizeWSDs += crt_len;
	    continue;
	  }

	int done = 0;
	
	while (!done)
	  {
	    
	    // optimization: check if S[diff[crt_pos]] and crtWSD are
	    // mutex. if yes, jump over entire S[diff[crt_pos]]
	    if (pos[crt_pos] == 0 && wsd_mutex(crtWSD, crt_len, S[diff[crt_pos]]->data, WSD_LEN)) {
	      if (crt_pos < dlen-1) {
		pos[crt_pos] = WSD_LEN;
		crt_pos++;
	      }
	      else {
		// compute probability of crtWSD
		double p = probability(crtWSD, crt_len);
		prob += p;
		//countWSDs++; sizeWSDs += crt_len;
		crt_pos--;
		//cout << "\t compute prob now" << endl;
	      }
	      continue;
	    }
	    
	    // the current wsd in diff is exhausted
	    if (pos[crt_pos] == WSD_LEN) {
	      
	      pos[crt_pos] = 0;
	      
	      /* optimization: we look at the last maps from the tail of the wsd */
	      crt_len = remove_maps_host (crtWSD, crt_len, diff[crt_pos]);	    
	      //print_maps(crtWSD,crt_len); cin.get();

	      crt_pos--;
	      if(crt_pos == -1)
		done = 1;
	      continue;
	    }
	    
	    Map *m1, *m2;
	    int var_count;
	    
	    do {
	      m1 = S[diff[crt_pos]]->data[pos[crt_pos]];
	      
	      pos[crt_pos]++;
	      
	      /* optimization: jump over those maps equal with an
	       * existing map in crtWSD
	       */
	      m2 = find_equal_map(crtWSD, crt_len, m1); 
	      
	      /* optimization: m1 is interesting only if P(neg(m1)\cup
	       * crtWSD) != 0. ie: 1+count # neg maps in crtWSD with
	       * m1->var != Rng[m1->var]
	       */
	      var_count = count_negvar (crtWSD, crt_len, m1->var)+1;	      
	    } while ((var_count == Rng[m1->var] || m2) &&  pos[crt_pos] < WSD_LEN);

	    /* failed search for extensions -> nothing to be generated */
	    if (pos[crt_pos] == WSD_LEN && (m2 || var_count == Rng[m1->var])) {
	      pos[crt_pos] = 0;
	      
	      /* optimization: we look at the last maps from the tail of the wsd */
	      crt_len = remove_maps_host (crtWSD, crt_len, diff[crt_pos]);
	      //print_maps(crtWSD,crt_len); cin.get();

	      crt_pos--;
	      if(crt_pos == -1)
		done = 1;
	      continue;
	    }	    
	    
	    if (!m2 && var_count < Rng[m1->var]) {
	      
	      /* fetch the last map in crtWSD; assumes crtWSD is not empty
	       * negate it, if it is of diff[crt_pos] 
	       */
	      if (crtWSD[crt_len-1]->wsd_host == diff[crt_pos])
		crtWSD[crt_len-1]->neg = 1 - crtWSD[crt_len-1]->neg;
	      
	      /*
	       * add copy_neg_map(m1) to crtWSD
	       */

	      if (crt_len == MAX_LEN) {
		MAX_LEN = 3*MAX_LEN/2;
		Map** tmp;
		if (tmp = (Map**)realloc(crtWSD, MAX_LEN * sizeof(Map*)))
		  crtWSD = tmp;
		else {
		  cout << "Reallocation error" << endl; exit(-1);
		}
	      }

	      crtWSD[crt_len] = copy_neg_map(m1);
	      crt_len++;
	      //print_maps(crtWSD,crt_len); cin.get();
	    }
	    
	    if (crt_pos < dlen-1)
	      crt_pos++;
	    else {
	      // compute probability of crtWSD
	      double p = probability(crtWSD, crt_len);
	      prob += p;
	      //countWSDs++; sizeWSDs += crt_len;
	      //cout << "\t compute prob now" << endl;
	    }
	  }//endwhile (!done)
	
	//free the list crtWSD
	reset_wsd(crtWSD,crt_len);

      }//endif test_bit
      }//endfor k
      
  /* statistics: 
   * countWSDs (how many wsds were generated), 
   * sizeWSDs/countWSDs (avg size of generated wsds)
   */

  return prob;
}

/***************** naive probability computation *************************/
/* iterate over all worlds */
double
naive_compute_prob ()
{
  int crt_rng[NUM_VARS];
  Map** world = (Map**)malloc(NUM_VARS*sizeof(Map*));

  double prob_world = 1.0;

  for (int i = 0; i < NUM_VARS; i++) {
    crt_rng[i] = 0;
    world[i] = (Map*)malloc(sizeof(Map));
    world[i]->var = i;
    world[i]->rng = 0;
    world[i]->neg = 0;
    prob_world = prob_world * (1.0/Rng[i]);
  }

  int finished = 0;
  double prob = 0.0;

  while (!finished)
    {
      int j = 0;
      while (j < NUM_WSDS && wsd_mutex(S[j]->data, WSD_LEN, world, NUM_VARS))
	j++;
      if (j < NUM_WSDS) //world accepted
	prob += prob_world;

      //go to the next world
      if (crt_rng[NUM_VARS-1] < Rng[NUM_VARS-1]-1) {
	crt_rng[NUM_VARS-1]++;
      } else {
	  int j = NUM_VARS-1;
	  int full_changed = 0;
	  while (j > -1 && !full_changed && !finished) {
	    if (crt_rng[j] == Rng[j]-1) {
	      crt_rng[j] = 0;
	      j--;
	    }
	    else {
	      crt_rng[j]++;
	      full_changed = 1;
	    }
	  }
	  if (j == -1) //all worlds are exhausted
	    finished = 1;
      }
      //update world
      for (int j = 0; j < NUM_VARS; j++)
	world[j]->rng = crt_rng[j];

    }
  return prob;
}



void
process_options (int count, char **vector)
{
  int option;
  
  while ((option = getopt (count, vector, "w:v:r:s:a:h:g:p::t::")) != -1)
    switch (option)
      {
      case 'w': //number of input world-set descriptors
	NUM_WSDS = atoi (optarg);
	break;
      case 'v': //number of variables
	NUM_VARS = atoi (optarg);
	break;
      case 'r': //range for variables
	RNG_MAX = atoi (optarg);
	if (RNG_MAX < 2) {
	  fprintf (stderr, "Max domain size of variables must be above 1\n");
	  exit(-1);
	}
	break;
      case 's': //size of wsds
	WSD_LEN = atoi (optarg);
	if (WSD_LEN < 2) {
	  fprintf (stderr, "Min size of world-set descriptors must be above 1\n");
	  exit(-1);
	}
	break;
      case 'a': // which algorithm: naive, ve, indve, we
	WHICH_ALGO = atoi (optarg);
	if (WHICH_ALGO < 0 || WHICH_ALGO > 3) {
	  fprintf (stderr, "Choose an implemented algo: 0 (naive), 1 (ve), 2 (indve), 3 (we)\n");
	  exit(-1);
	}
	break;
      case 'h': // which heuristic for ve: minlog, minmax
	WHICH_HEURISTIC = atoi (optarg);
	if (WHICH_HEURISTIC < 0 || WHICH_HEURISTIC > 1) {
	  fprintf (stderr, "Choose an heuristic for variable elimination : 0 (minlog), 1 (minmax)\n");
	  exit(-1);
	}
	break;
      case 'g': // which data generation: 0 (random), 1 (random from joined BIDs), 2 (random from self-joined BIDs)
	//choice 0 is now inactive
	DATA_GEN = atoi (optarg);
	if (DATA_GEN < 1 || DATA_GEN > 2) {
	  fprintf (stderr, "Choose an implemented data generation algo: 1 (random,joined BIDs), 2 (random, self-joined BIDs)\n");
	  exit(-1);
	}
	break;
      case 'p'://optional; print data and exit.
	PRINT_MODE = 1;
	break;
      case 't'://optional; test mode: does not print the conf value
	TEST_MODE = 1;
	break;
      default:
	printf("Unknown command line option..\n");
	printf("Try: %s [-w #wsds] [-v #vars] [-r rng-vars] [-s wsd size] [-g datagen] [-a algo] [-h heuristic] [-t] [-p] \n",vector[0]);
	exit(-1);
      }
  if (!TEST_MODE) {
    fprintf (stderr, "TEST_MODE (0=no, 1=yes): 0\n");    
    fprintf (stderr, "PRINT_MODE (0=no, 1=yes): %i\n",PRINT_MODE);    
    fprintf (stderr, "Number of variables: %i\n",NUM_VARS);
    fprintf (stderr, "Max domain size of variables: %i\n",RNG_MAX);
    fprintf (stderr, "Number of ws-descriptors: %i\n",NUM_WSDS);    
    fprintf (stderr, "Size of world-set descriptors: %i\n",WSD_LEN);
    fprintf (stderr, "Data generation algorithm (1=random+joinedBIDs, 2=random+selfjoinedBIDS): %i\n",DATA_GEN);
    fprintf (stderr, "Computation algorithm (0=naive, 1=ve, 2=indve, 3=we): %i\n",WHICH_ALGO);
    if (WHICH_ALGO == 1 || WHICH_ALGO == 2)
      fprintf (stderr, "Heuristic for variable elimination: 0 (minlog), 1 (minmax): %i\n",WHICH_HEURISTIC);
  }
  return;
}

int main(int ac, char **av)
{
  //generate time-dependent data
  srand(time(NULL));

  //used for checking time usage for data generation
  //  clock_t t_gendata_start, t_gendata_end;


  //default values
  NUM_WSDS = 10;
  NUM_VARS = 10;
  RNG_MAX  = 10;
  TEST_MODE = 0;
  PRINT_MODE = 0;
  WSD_LEN = 4;
  WHICH_ALGO = 2;
  WHICH_HEURISTIC = 0;
  DATA_GEN = 1;

  process_options (ac, av);

  //t_gendata_start = clock(); 

  Rng = (int*)malloc(NUM_VARS*sizeof(int));
  create_world_table();
  create_var_rng_table();

  switch (DATA_GEN) {
  case 0: // random
    create_random_WSDs();
    break;
  case 1: //random, joined BIDs
    create_random_WSDs_from_BIDs();
    break;
  case 2: //random, self-joined BIDs
    create_WSDs_from_selfjoined_BIDs(0.10);
    break;
  }

  //t_gendata_end = clock(); 

  //  if (PRINT_MODE)
  //exit(1);

  bitset* set = bitset_init(NUM_WSDS);
  bitset_set(set);

  double prob = 0.0;

  switch (WHICH_ALGO) {
  case 0: //naive
    prob = naive_compute_prob();
    break;
  case 1: //ve
    prob = ve_compute_prob(set);
    break;
  case 2: //indve 
    prob = indve_compute_prob(set);
    break;
  case 3: //we
    prob = we_compute_prob(set);
    break;
  }

  if (!TEST_MODE) {
    cout << prob << endl;
    //cout << "Time needed to generate data:" << (t_gendata_end - t_gendata_start)/CLOCKS_PER_SEC << endl;
  } //else
    //cout << (t_gendata_end - t_gendata_start)/CLOCKS_PER_SEC << "\t";

  bitset_free(set);
  free(Rng);
  //free S
}
