
// The Karp/Luby FPRAS for approximately counting the number of satisfying
// truth assignments of a DNF.  Adapted to compute confidence values.

// Compile with g++ option -O3, it makes a big difference (because of STL).



using namespace std;

#include <iostream>
#include <math.h>
#include <stdlib.h>
#include <vector>



// default parameter values.
//
int NUM_CLAUSES    = 10;   // number of clauses in the DNF
int NUM_VARIABLES  = 5;    // number of propositional variables
int MAX_RANGE      = 2;    // number of alternatives per variables
int CLAUSE_SIZE    = 2;    // number of literals (var. assignments) per clause
double epsilon     = 0.01;
const double delta = 0.01;

int TEST_MODE = 0; // suppress output? FIXME: This should be a #define.


// for each clause, (variable id, positive/negative)-pairs
vector<pair<int,int> > * Cv;

int*     r;               //[NUM_CLAUSES];     // #vars in clause
double   nM = 0;          //sum of the confidences of all clauses
int*     range;           //[NUM_VARIABLES];
double** probs;           //[NUM_VARIABLES][MAX_RANGE];
double*  clause_bag_prob; //[NUM_CLAUSES];



// process command-line options.
void process_options (int count, char **vector)
{
  int option;
  
  while ((option = getopt (count, vector, "e:w:s:v:r:t::")) != -1)
    switch (option)
      {
      case 'e': //number of input world-set descriptors
	epsilon = atof (optarg);
	if (epsilon <= 0.0 || epsilon >= 1) {
	  cerr << "Allowed epsilon in (0,1)" << endl;
	  exit(-1);
	}
	break;
      case 'w': 
	NUM_CLAUSES = atoi (optarg);
	break;
      case 's': 
	CLAUSE_SIZE = atoi (optarg);
	if (CLAUSE_SIZE < 2) {
	  cerr << "Min size of clauses must be above 1" << endl;
	  exit(-1);
	}
	break;
      case 'v': 
	NUM_VARIABLES = atoi (optarg);
	break;
      case 'r': 
	MAX_RANGE = atoi (optarg);
	if (MAX_RANGE < 2) {
	  cerr << "Max domain size of variables must be above 2" << endl;
	  exit(-1);
	}
	break;
      case 't'://optional; test mode: does not print the conf value
	TEST_MODE = 1;
	break;
      default:
	cerr << "Unknown command line option.." << endl;
	cerr << "USAGE: " << vector[0]
             << " [-w #wsds] [-v #vars] [-r rng-vars] [-s wsd size] [-g datagen] [-a algo] [-h heuristic] [-t] [-p]" << endl;
	exit(-1);
      }
  if (!TEST_MODE) {
    cerr << "TEST_MODE (0=no, 1=yes): 0" << endl;
    cerr << "Number of variables: " << NUM_VARIABLES << endl;
    cerr << "Max domain size of variables: " << MAX_RANGE << endl;
    cerr << "Number of ws-descriptors: " << NUM_CLAUSES << endl;
    cerr << "Size of world-set descriptors: " << CLAUSE_SIZE << endl;
    cerr << "epsilon: " << epsilon << endl;
  }
}


// Choose an index i from [0..distrib_size] with probability distrib[i]
// This could be optimized by performing binary search.
// if distrib is a null pointer, we assume a uniform distribution by default.
int choose_with_distribution(int distrib_size, double *distrib)
{
  if(distrib == 0) // uniform distribution
	  return rand() % distrib_size;

  int i;
  double pick = (double)rand(); // in interval [0 .. RAND_MAX]
  double offset = 0;

  for(i = 0; i < distrib_size; i++)
  {
    offset += distrib[i] * (double)RAND_MAX;	 
    if(pick <= offset) break;
  }
  return i;
}



// returns an unbiased estimator for the probability of the DNF.
// This is the version from the Vazirani book and converges more
// quickly than the basic coverage algorithm from Karp-Luby-Madras.
// It is similar to the algorithm from section 5 of that paper.
// 
double compute_estimator()
{
    int i = choose_with_distribution(NUM_CLAUSES, clause_bag_prob);

    // pick one truth assignment tau from those of C[i] with probability
    // P(tau)/Sum_{tau' in C[i]} P(tau').
    //
    int tau[NUM_VARIABLES];
    for(unsigned j = 0; j < NUM_VARIABLES; j++)
    {
      int Cij = -1;
      for(unsigned k = 0; k < Cv[i].size(); k++)
         if(Cv[i][k].first == j) Cij = Cv[i][k].second;

      if(Cij == -1) // clause i does not determine value of x_j
      {
        // randomly complete the truth assignment with the right weight.
        // currently the local worlds are uniformly distributed.

	tau[j] = choose_with_distribution(range[j], probs[j]);
      }
      else tau[j] = Cij;
    }
    
    // Here is where the difference to the basic Karp-Luby estimator
    // starts. Rather than checking whether the chosen clause is the first
    // among those that satisfy the chosen possible world with respect to
    // some arbitrary fixed order of the clauses, and returning either
    // 1 or 0 based on this, we count the number of clauses that satisfy
    // the world and return the ratio of nM to that count.
 
    // c_tau: count how many clauses are satisfied by tau
    int c_tau = 0;
    for(i = 0; i < NUM_CLAUSES; i++)
    {
      bool clause_satisfied = true;
      for(unsigned k = 0; k < Cv[i].size(); k++)
        if(tau[Cv[i][k].first] != Cv[i][k].second)
        {
          clause_satisfied = false;
          break;
        }
      if(clause_satisfied) c_tau ++;
    }

    // return an unbiased estimator for the probability.
    return (double)nM / (double)c_tau;
}



const double e = 2.718281828459;


// From Dagum, Karp, Luby, Ross, "An Optimal Algorithm for
// Monte Carlo Estimation".
// This is an auxiliary procedure of the AA_algorithm().
// Do not call directly.
double stopping_rule_algorithm(double epsilon, double delta)
{
   const double upsilon  = 4.0 * (e - 2.0) * log(2.0 / delta)
                         / (epsilon * epsilon);
   const double upsilon1 = 1.0 + (1.0 + epsilon) * upsilon;

   int    N = 0;
   double S = 0;

   while(S < upsilon1)
   {
      N++;
      S += compute_estimator();
   }

   return upsilon1 / (double)N;
}



// from Dagum, Karp, Luby, Ross, "An Optimal Algorithm for
// Monte Carlo Estimation".
double AA_algorithm()
{
   // is log the natural logarithm?
   const double upsilon  = 4.0 * (e - 2.0) * log(2/delta) / (epsilon * epsilon);
   const double upsilon2 = 2.0 * (1.0 + sqrt(epsilon))
                         * (1.0 + 2.0 * sqrt(epsilon))
                         * (1.0 + log(1.5))/log(2/delta) * upsilon;

   // Step 1:
   double mu_hat =
      stopping_rule_algorithm(min(0.5, sqrt(epsilon)), delta/3);

   // Step 2:
   const double N = ceil(upsilon2 * epsilon / mu_hat);
   double S = 0;
   for(unsigned i = 1; i <= N; i++)
   {
                // this is not 2 * compute_estimator(), of course.
      double kl2 = compute_estimator() + compute_estimator();

      S += kl2 * kl2 / 2.0;
   }
   double rho_hat = max(S / N, epsilon * mu_hat);

   // Step 3:
   double N2 = upsilon2 * rho_hat / (mu_hat * mu_hat);

   cout << "Number of iterations (AA alg.) = " << N2 << endl;

   S = 0;
   for(unsigned i = 0; i < N2; i++)
       S += compute_estimator();

   return S / N2;
}


// Sets up the data structures that represent the probability distributitions
// (currently uniform, with range[i] alternatives for the i-th variable),
// clauses (C, r), and associated clause weights (clause_bag_prob).
//
void init_data_structures()
{
/////// Allocate memory for our data structures that hold the DNF
/////// as well as the probability distributions of the discrete random
/////// variables.

  range = (int*)malloc(NUM_VARIABLES * sizeof(int));

  Cv = new vector<pair<int,int> >[NUM_CLAUSES];

  probs = new double*[NUM_VARIABLES];
  for(unsigned i = 0; i < NUM_VARIABLES; i++)
     probs[i] = new double[MAX_RANGE];

  // #vars in clause
  r = new int[NUM_CLAUSES];

  // confidence of wsd C[i]
  double nS[NUM_CLAUSES];

  // nM: sum of the confidences of all wsds
  nM = 0;

  // number of alternatives of variable.
  range = (int*)malloc(NUM_VARIABLES * sizeof(int));

  clause_bag_prob = new double[NUM_CLAUSES];


////// The code below is specific to our VLDB 2008 experiments.


  // number of local worlds per variable
  for(unsigned i = 0; i < NUM_VARIABLES; i++)
  {
    // To do: generalize. Currently, in this set of experiments, 
    // all variables have the maximum range and use a uniform distribution.

    range[i] = MAX_RANGE;

    for(unsigned j = 0; j < range[i]; j++)
       probs[i][j] = 1.0 / MAX_RANGE;
  }


  int var_partitions[NUM_CLAUSES];
  int count = 0;


  for (int i = 0; i < CLAUSE_SIZE - 1; i++)
  {
    var_partitions[i] = 1 + (int)(NUM_VARIABLES * (rand()
                      / (CLAUSE_SIZE * (RAND_MAX + 1.0))));
    count += var_partitions[i];
  }
  var_partitions[CLAUSE_SIZE-1] = NUM_VARIABLES - count;



  for (int i = 0; i < NUM_CLAUSES; i++)
  {
    r[i] = CLAUSE_SIZE;

    count = 0;
    for (int j = 0; j < CLAUSE_SIZE; j++)
    {
      int var = count + (int) (var_partitions[j] * (rand() / (RAND_MAX + 1.0)));
      count += var_partitions[j];
      Cv[i].push_back(pair<int,int>(var, rand() % range[var]));
    }
      
    nS[i] = pow(1.0 / MAX_RANGE, r[i]); // confidence of clause i
    nM += nS[i];
  }
  for (int i = 0; i < NUM_CLAUSES; i++)
  {
    clause_bag_prob[i] = nS[i] / nM;
  }
}



double basic_kl()
{
  // compute the average of the results of many invocations of the
  // unbiased estimator computation.

  // Number of iterations required to ensure error guarantees
  int n_loops = (int)ceil(4.0 * NUM_CLAUSES * log(2.0/delta)
                       / (epsilon * epsilon));

  cout << "Number of Iterations (basic KL): " << n_loops << endl;

  double sum = 0;
  for(int rep = 0; rep < n_loops; rep++)
  {
    sum += compute_estimator();
  }

  // return the average of the unbiased estimators.
  return sum / n_loops;
}



int main(int ac, char **av)
{
  //generate time-dependent data
  srand(time(NULL));

  process_options (ac, av);
  init_data_structures();
  cout << "Data structures ready." << endl;

  // output clauses
  for(unsigned i = 0; i < NUM_CLAUSES; i++)
  {
     for(unsigned j = 0; j < Cv[i].size(); j++)
     {
        cout << "x" << Cv[i][j].first << "=" << Cv[i][j].second << " ";
     }
     cout << endl;
  }


  double y = basic_kl();
  if(!TEST_MODE) cout << y << endl;

  double x = AA_algorithm();
  if(!TEST_MODE) cout << x << endl;
}


